@@ -1,8 +1,6 @@
-from contextlib import contextmanager
-from warnings import warn
 import inspect
-from .dispatcher import Dispatcher, MethodDispatcher, ambiguity_warn
 
+from .dispatcher import Dispatcher, MethodDispatcher, ambiguity_warn
 
 global_namespace = dict()
 
@@ -51,21 +49,23 @@ def dispatch(*types, **kwargs):
     ...         self.data = [datum]
     """
     namespace = kwargs.get('namespace', global_namespace)
-    on_ambiguity = kwargs.get('on_ambiguity', ambiguity_warn)
 
     types = tuple(types)
+
     def _(func):
         name = func.__name__
 
         if ismethod(func):
-            dispatcher = inspect.currentframe().f_back.f_locals.get(name,
-                MethodDispatcher(name))
+            dispatcher = inspect.currentframe().f_back.f_locals.get(
+                name,
+                MethodDispatcher(name),
+            )
         else:
             if name not in namespace:
                 namespace[name] = Dispatcher(name)
             dispatcher = namespace[name]
 
-        dispatcher.add(types, func, on_ambiguity=on_ambiguity)
+        dispatcher.add(types, func)
         return dispatcher
     return _
 
@@ -76,5 +76,9 @@ def ismethod(func):
     Note that this has to work as the method is defined but before the class is
     defined.  At this stage methods look like functions.
     """
-    spec = inspect.getargspec(func)
-    return spec and spec.args and spec.args[0] == 'self'
+    if hasattr(inspect, "signature"):
+        signature = inspect.signature(func)
+        return signature.parameters.get('self', None) != None
+    else:
+        spec = inspect.getargspec(func)
+        return spec and spec.args and spec.args[0] == 'self'
