@@ -1,6 +1,10 @@
 from warnings import warn
+import inspect
 from .conflict import ordering, ambiguities, super_signature, AmbiguityWarning
 from .utils import expand_tuples
+import itertools as itl
+
+
 
 class MDNotImplementedError(NotImplementedError):
     """ A NotImplementedError for multiple dispatch """
@@ -23,21 +27,24 @@ def ambiguity_warn(dispatcher, ambiguities):
     warn(warning_text(dispatcher.name, ambiguities), AmbiguityWarning)
 
 
-_unresolved_dispatchers = set()
-_resolve = [True]
-
-
 def halt_ordering():
-    _resolve[0] = False
+    """Deprecated interface to temporarily disable ordering.
+    """
+    warn(
+        'halt_ordering is deprecated, you can safely remove this call.',
+        DeprecationWarning,
+    )
 
 
 def restart_ordering(on_ambiguity=ambiguity_warn):
-    _resolve[0] = True
-    while _unresolved_dispatchers:
-        dispatcher = _unresolved_dispatchers.pop()
-        dispatcher.reorder(on_ambiguity=on_ambiguity)
-
-
+    """Deprecated interface to temporarily resume ordering.
+    """
+    warn(
+        'restart_ordering is deprecated, if you would like to eagerly order'
+        'the dispatchers, you should call the ``reorder()`` method on each'
+        ' dispatcher.',
+        DeprecationWarning,
+    )
 
 class Dispatcher(object):
     """ Dispatch methods based on type signature
@@ -61,15 +68,15 @@ class Dispatcher(object):
     >>> f(3.0)
     2.0
     """
-    __slots__ = 'name', 'funcs', 'ordering', '_cache', 'doc'
+    __slots__ = '__name__', 'name', 'funcs', '_ordering', '_cache', 'doc'
 
     def __init__(self, name, doc=None):
-        self.name = name
-        self.funcs = dict()
-        self._cache = dict()
-        self.ordering = []
+        self.name = self.__name__ = name
+        self.funcs = {}
         self.doc = doc
 
+        self._cache = {}
+
     def register(self, *types, **kwargs):
         """ register dispatcher with new implementation
 
@@ -101,7 +108,33 @@ class Dispatcher(object):
             return func
         return _
 
-    def add(self, signature, func, on_ambiguity=ambiguity_warn):
+    @classmethod
+    def get_func_params(cls, func):
+        if hasattr(inspect, "signature"):
+            sig = inspect.signature(func)
+            return sig.parameters.values()
+
+    @classmethod
+    def get_func_annotations(cls, func):
+        """ get annotations of function positional parameters
+        """
+        params = cls.get_func_params(func)
+        if params:
+            Parameter = inspect.Parameter
+
+            params = (param for param in params
+                      if param.kind in
+                      (Parameter.POSITIONAL_ONLY,
+                       Parameter.POSITIONAL_OR_KEYWORD))
+
+            annotations = tuple(
+                param.annotation
+                for param in params)
+
+            if all(ann is not Parameter.empty for ann in annotations):
+                return annotations
+
+    def add(self, signature, func):
         """ Add new types/method pair to dispatcher
 
         >>> D = Dispatcher('add')
@@ -119,34 +152,48 @@ class Dispatcher(object):
         with a dispatcher/itself, and a set of ambiguous type signature pairs
         as inputs.  See ``ambiguity_warn`` for an example.
         """
+        # Handle annotations
+        if not signature:
+            annotations = self.get_func_annotations(func)
+            if annotations:
+                signature = annotations
+
         # Handle union types
         if any(isinstance(typ, tuple) for typ in signature):
             for typs in expand_tuples(signature):
-                self.add(typs, func, on_ambiguity)
+                self.add(typs, func)
             return
 
         for typ in signature:
             if not isinstance(typ, type):
                 str_sig = ', '.join(c.__name__ if isinstance(c, type)
-                                               else str(c) for c in signature)
+                                    else str(c) for c in signature)
                 raise TypeError("Tried to dispatch on non-type: %s\n"
                                 "In signature: <%s>\n"
                                 "In function: %s" %
                                 (typ, str_sig, self.name))
 
         self.funcs[signature] = func
-        self.reorder(on_ambiguity=on_ambiguity)
         self._cache.clear()
 
-    def reorder(self, on_ambiguity=ambiguity_warn):
-        if _resolve[0]:
-            self.ordering = ordering(self.funcs)
-            amb = ambiguities(self.funcs)
-            if amb:
-                on_ambiguity(self, amb)
-        else:
-            _unresolved_dispatchers.add(self)
+        try:
+            del self._ordering
+        except AttributeError:
+            pass
 
+    @property
+    def ordering(self):
+        try:
+            return self._ordering
+        except AttributeError:
+            return self.reorder()
+
+    def reorder(self, on_ambiguity=ambiguity_warn):
+        self._ordering = od = ordering(self.funcs)
+        amb = ambiguities(self.funcs)
+        if amb:
+            on_ambiguity(self, amb)
+        return od
 
     def __call__(self, *args, **kwargs):
         types = tuple([type(arg) for arg in args])
@@ -156,25 +203,27 @@ class Dispatcher(object):
             func = self.dispatch(*types)
             if not func:
                 raise NotImplementedError(
-                        'Could not find signature for %s: <%s>' %
-                        (self.name, str_signature(types)))
+                    'Could not find signature for %s: <%s>' %
+                    (self.name, str_signature(types)))
             self._cache[types] = func
         try:
             return func(*args, **kwargs)
 
         except MDNotImplementedError:
             funcs = self.dispatch_iter(*types)
-            next(funcs) # burn first
+            next(funcs)  # burn first
             for func in funcs:
                 try:
                     return func(*args, **kwargs)
                 except MDNotImplementedError:
                     pass
-            raise NotImplementedError("Matching functions for "
-                    "%s: <%s> found, but none completed successfully"
-                    % (self.name, str_signature(types)))
-
 
+            raise NotImplementedError(
+                "Matching functions for "
+                "%s: <%s> found, but none completed successfully" % (
+                    self.name, str_signature(types),
+                ),
+            )
 
     def __str__(self):
         return "<dispatched %s>" % self.name
@@ -235,10 +284,9 @@ class Dispatcher(object):
     def __setstate__(self, d):
         self.name = d['name']
         self.funcs = d['funcs']
-        self.ordering = ordering(self.funcs)
+        self._ordering = ordering(self.funcs)
         self._cache = dict()
 
-
     @property
     def __doc__(self):
         docs = ["Multiply dispatched method: %s" % self.name]
@@ -262,6 +310,29 @@ class Dispatcher(object):
 
         return '\n\n'.join(docs)
 
+    def _help(self, *args):
+        return self.dispatch(*map(type, args)).__doc__
+
+    def help(self, *args, **kwargs):
+        """ Print docstring for the function corresponding to inputs """
+        print(self._help(*args))
+
+    def _source(self, *args):
+        func = self.dispatch(*map(type, args))
+        if not func:
+            raise TypeError("No function found")
+        return source(func)
+
+    def source(self, *args, **kwargs):
+        """ Print source code for the function corresponding to inputs """
+        print(self._source(*args))
+
+
+def source(func):
+    s = 'File: %s\n\n' % inspect.getsourcefile(func)
+    s = s + inspect.getsource(func)
+    return s
+
 
 class MethodDispatcher(Dispatcher):
     """ Dispatch methods based on type signature
@@ -269,6 +340,14 @@ class MethodDispatcher(Dispatcher):
     See Also:
         Dispatcher
     """
+    __slots__ = ('obj', 'cls')
+
+    @classmethod
+    def get_func_params(cls, func):
+        if hasattr(inspect, "signature"):
+            sig = inspect.signature(func)
+            return itl.islice(sig.parameters.values(), 1, None)
+
     def __get__(self, instance, owner):
         self.obj = instance
         self.cls = owner
@@ -291,13 +370,15 @@ def str_signature(sig):
     """
     return ', '.join(cls.__name__ for cls in sig)
 
+
 def warning_text(name, amb):
     """ The text for ambiguity warnings """
-    text = "\nAmbiguities exist in dispatched function %s\n\n"%(name)
+    text = "\nAmbiguities exist in dispatched function %s\n\n" % (name)
     text += "The following signatures may result in ambiguous behavior:\n"
     for pair in amb:
-        text += "\t" + ', '.join('['+str_signature(s)+']' for s in pair) + "\n"
+        text += "\t" + \
+            ', '.join('[' + str_signature(s) + ']' for s in pair) + "\n"
     text += "\n\nConsider making the following additions:\n\n"
     text += '\n\n'.join(['@dispatch(' + str_signature(super_signature(s))
-                      + ')\ndef %s(...)'%name for s in amb])
+                         + ')\ndef %s(...)' % name for s in amb])
     return text
